@@ -2,9 +2,10 @@
 Utility functions for teaching data science.
 """
 
-from typing import Iterable, List, Tuple
+from typing import Dict, Iterable, List, Tuple
 
 import re
+import os
 import numpy
 import statistics
 import matplotlib
@@ -19,6 +20,25 @@ import math
 from data_algebra.cdata import RecordMap, RecordSpecification
 
 
+def types_in_frame(d: pandas.DataFrame) -> Dict[str, List[type]]:
+    """
+    Report what type as seen as values in a Pandas data frame.
+    
+    :param d: Pandas data frame to inspect, not altered.
+    :return: dictionary mapping column names to order lists of types found in column.
+    """
+    assert isinstance(d, pandas.DataFrame)
+    type_dict_map = {
+        col_name: {str(type(v)): type(v) for v in d[col_name]}
+            for col_name in d.columns
+    }
+    type_dict = {
+        col_name: [type_set[k] for k in sorted(list(type_set.keys()))]
+            for col_name, type_set in type_dict_map.items()
+    }
+    return type_dict
+
+
 # noinspection PyPep8Naming
 def cross_predict_model(
     fitter, X: pandas.DataFrame, y: pandas.Series, plan: List
@@ -38,8 +58,8 @@ def cross_predict_model(
     assert isinstance(plan, List)
     preds = None
     for pi in plan:
-        model = fitter.fit(X.iloc[pi["train"]], y.iloc[pi["train"]])
-        predg = model.predict(X.iloc[pi["test"]])
+        model = fitter.fit(X.iloc[pi["train"], :], y.iloc[pi["train"]])
+        predg = model.predict(X.iloc[pi["test"], :])
         # patch results in
         if preds is None:
             preds = numpy.asarray([None] * X.shape[0], dtype=numpy.asarray(predg).dtype)
@@ -66,8 +86,8 @@ def cross_predict_model_proba(
     assert isinstance(plan, List)
     preds = None
     for pi in plan:
-        model = fitter.fit(X.iloc[pi["train"]], y.iloc[pi["train"]])
-        predg = model.predict_proba(X.iloc[pi["test"]])
+        model = fitter.fit(X.iloc[pi["train"], :], y.iloc[pi["train"]])
+        predg = model.predict_proba(X.iloc[pi["test"], :])
         # patch results in
         if preds is None:
             preds = numpy.zeros((X.shape[0], predg.shape[1]))
@@ -898,17 +918,7 @@ def fit_onehot_enc(
         categories="auto", drop=None, sparse=False, handle_unknown="ignore"  # default
     )
     enc.fit(d[categorical_var_names])
-    produced_column_names = list(enc.get_feature_names())
-    # map back to original column names
-
-    def replace_col_name(v):
-        """Replace x[0-9]+_level with var_level"""
-        v_prefix = re.sub(r"_.*$", "", v)
-        v_suffix = re.sub(r"^x[0-9]+_", "", v)
-        v_index = int(re.sub(r"^x", "", v_prefix))
-        return f"{categorical_var_names[v_index]}_{v_suffix}"
-
-    produced_column_names = [replace_col_name(v) for v in produced_column_names]
+    produced_column_names = list(enc.get_feature_names_out())
     # return the structure
     encoder_bundle = {
         "categorical_var_names": categorical_var_names,
@@ -938,3 +948,35 @@ def apply_onehot_enc(d: pandas.DataFrame, *, encoder_bundle: dict) -> pandas.Dat
     complementary_columns = [c for c in d.columns if c not in cat_set]
     res = pandas.concat([d[complementary_columns], one_hotted], axis=1)
     return res
+
+
+# https://stackoverflow.com/a/56695622/6901725
+# from https://stackoverflow.com/questions/11130156/suppress-stdout-stderr-print-from-python-functions
+class suppress_stdout_stderr(object):
+    '''
+    A context manager for doing a "deep suppression" of stdout and stderr in
+    Python, i.e. will suppress all print, even if the print originates in a
+    compiled C/Fortran sub-function.
+       This will not suppress raised exceptions, since exceptions are printed
+    to stderr just before a script exits, and after the context manager has
+    exited (at least, I think that is why it lets exceptions through).
+
+    '''
+    def __init__(self):
+        # Open a pair of null files
+        self.null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
+        # Save the actual stdout (1) and stderr (2) file descriptors.
+        self.save_fds = (os.dup(1), os.dup(2))
+
+    def __enter__(self):
+        # Assign the null pointers to stdout and stderr.
+        os.dup2(self.null_fds[0], 1)
+        os.dup2(self.null_fds[1], 2)
+
+    def __exit__(self, *_):
+        # Re-assign the real stdout/stderr back to (1) and (2)
+        os.dup2(self.save_fds[0], 1)
+        os.dup2(self.save_fds[1], 2)
+        # Close the null files
+        os.close(self.null_fds[0])
+        os.close(self.null_fds[1])
